/*
 * Copyright (C) 2008 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file adu.h \brief Global definitions. */

#include <sys/stat.h>
#include <fcntl.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h> /* time(), localtime() */
#include <unistd.h>
#include <errno.h>
#include <limits.h>
#include <stdarg.h>
#include <inttypes.h>
#include <string.h>
#include <assert.h>
#include <osl.h>
#include "gcc-compat.h"
#include "portable_io.h"

/** debug loglevel, gets really noisy */
#define DEBUG 1
/** still noisy, but won't fill your disk */
#define INFO  2
/** normal, but significant event */
#define NOTICE 3
/** unexpected event that can be handled */
#define WARNING 4
/** unhandled error condition */
#define ERROR 5
/** system might be unreliable */
#define CRIT 6
/** last message before exit */
#define EMERG 7

/** Log messages with lower priority than that will not be compiled in. */
#define COMPILE_TIME_LOGLEVEL 0

/** \cond */
#if DEBUG > COMPILE_TIME_LOGLEVEL
#define DEBUG_LOG(f,...) __log(DEBUG, "%s: " f, __FUNCTION__, ## __VA_ARGS__)
#else
#define DEBUG_LOG(...) do {;} while (0)
#endif

#if INFO > COMPILE_TIME_LOGLEVEL
#define INFO_LOG(f,...) __log(INFO, "%s: " f, __FUNCTION__, ## __VA_ARGS__)
#else
#define INFO_LOG(...) do {;} while (0)
#endif

#if NOTICE > COMPILE_TIME_LOGLEVEL
#define NOTICE_LOG(f,...) __log(NOTICE, "%s: " f, __FUNCTION__, ## __VA_ARGS__)
#else
#define NOTICE_LOG(...) do {;} while (0)
#endif

#if WARNING > COMPILE_TIME_LOGLEVEL
#define WARNING_LOG(f,...) __log(WARNING, "%s: " f, __FUNCTION__, ##  __VA_ARGS__)
#else
#define WARNING_LOG(...) do {;} while (0)
#endif

#if ERROR > COMPILE_TIME_LOGLEVEL
#define ERROR_LOG(f,...) __log(ERROR, "%s: " f, __FUNCTION__, ## __VA_ARGS__)
#else
#define ERROR_LOG(...) do {;} while (0)
#endif

#if CRIT > COMPILE_TIME_LOGLEVEL
#define CRIT_LOG(f,...) __log(CRIT, "%s: " f, __FUNCTION__, ## __VA_ARGS__)
#else
#define CRIT_LOG(...) do {;} while (0)
#endif

#if EMERG > COMPILE_TIME_LOGLEVEL
#define EMERG_LOG(f,...) __log(EMERG, "%s: " f, __FUNCTION__, ## __VA_ARGS__)
#else
#define EMERG_LOG(...)
#endif
/** \endcond */

/**
 * Wrapper for isspace.
 * NetBSD needs this.
 */
/*
 * The values should be cast to an unsigned char first, then to int.
 * Why? Because the isdigit (as do all other is/to functions/macros)
 * expect a number from 0 upto and including 255 as their (int) argument.
 * Because char is signed on most systems, casting it to int immediately
 * gives the functions an argument between -128 and 127 (inclusive),
 * which they will use as an array index, and which will thus fail
 * horribly for characters which have their most significant bit set.
 */
#define adu_isspace(c) isspace((int)(unsigned char)(c))

/**
 * Write a log message to a dynamically allocated string.
 *
 * \param fmt Usual format string.
 * \param p Result pointer.
 *
 * \sa printf(3). */
#define VSPRINTF(fmt, p) \
{ \
	int n; \
	size_t size = 100; \
	p = adu_malloc(size); \
	while (1) { \
		va_list ap; \
		/* Try to print in the allocated space. */ \
		va_start(ap, fmt); \
		n = vsnprintf(p, size, fmt, ap); \
		va_end(ap); \
		/* If that worked, return the string. */ \
		if (n > -1 && n < size) \
			break; \
		/* Else try again with more space. */ \
		if (n > -1) /* glibc 2.1 */ \
			size = n + 1; /* precisely what is needed */ \
		else /* glibc 2.0 */ \
			size *= 2; /* twice the old size */ \
		p = adu_realloc(p, size); \
	} \
}

/** Evaluates to 1 if x < y, to -1 if x > y and to 0 if x == y. */
#define NUM_COMPARE(x, y) ((int)((x) < (y)) - (int)((x) > (y)))

/** The columns of the directory table. */
enum dir_table_columns {
	/** The name of the directory. */
	DT_NAME,
	/** The dir count number. */
	DT_NUM,
	/** The number of the parent directory. */
	DT_PARENT_NUM,
	/** The number of bytes of all regular files. */
	DT_BYTES,
	/** The number of all regular files. */
	DT_FILES,
	/** Number of columns in this table. */
	NUM_DT_COLUMNS
};

extern struct osl_table *dir_table;

/** The adu command line options. */
extern struct gengetopt_args_info conf;

/** Computed database dir. */
extern char *database_dir;

/**
 * The select command line options.
 *
 * Either given at the command line, or via the \a set command
 * in interactive mode.
 */
extern struct select_args_info select_conf;

/**
 * Compare two osl objects pointing to unsigned integers of 64 bit size.
 *
 * \param obj1 Pointer to the first integer.
 * \param obj2 Pointer to the second integer.
 *
 * \return The values required for an osl compare function.
 *
 * \sa osl_compare_func, osl_hash_compare().
 */
_static_inline_ int uint64_compare(const struct osl_object *obj1,
		const struct osl_object *obj2)
{
	uint64_t d1 = read_u64((const char *)obj1->data);
	uint64_t d2 = read_u64((const char *)obj2->data);

	if (d1 < d2)
		return 1;
	if (d1 > d2)
		return -1;
	return 0;
}

/**
 * Compare the size of two directories
 *
 * \param obj1 Pointer to the first object.
 * \param obj2 Pointer to the second object.
 *
 * This function first compares the size values as usual integers. If they compare as
 * equal, the address of \a obj1 and \a obj2 are compared. So this compare function
 * returns zero if and only if \a obj1 and \a obj2 point to the same memory area.
 */
_static_inline_ int size_compare(const struct osl_object *obj1, const struct osl_object *obj2)
{
	uint64_t d1 = *(uint64_t *)obj1->data;
	uint64_t d2 = *(uint64_t *)obj2->data;
	int ret = NUM_COMPARE(d2, d1);

	if (ret)
		return ret;
	//INFO_LOG("addresses: %p, %p\n", obj1->data, obj2->data);
	return NUM_COMPARE(obj2->data, obj1->data);
}

/* adu.c */
__printf_2_3 void __log(int, const char*, ...);
int open_dir_table(int create);
void check_signals(void);
/* create.c */
int com_create(void);

/* interactive.c */
void print_interactive_help(void);
int com_interactive(void);
